#include "VpnPeer.hpp"
#include "UdpVpn.hpp"

#include <cstdint>
#include <cstring>
#include <functional>

const double RTTLogger::EXP_AVG_FACTOR = 0.75;
const unsigned int RTTLogger::BASE_UPDATE_DELAY = 1000; // ms

VpnPeer::VpnPeer(UdpVpn* vpn, const sockaddr_in6& ext_addr,
        const in6_addr& int_addr)
    : _vpn(vpn), _ext_addr(ext_addr), _int_addr(int_addr), _next_send_seqno(0)
{
    cycle_next_control();
}

void VpnPeer::cycle_next_control() {
    _next_control_packet =
        std::make_unique<VpnControlPacket>(_vpn->get_mtu(), false);
    _next_control_packet->set_peer(this);
}

void VpnPeer::set_int_addr(const in6_addr& int_addr) {
    memcpy(&_int_addr, &int_addr, sizeof(_int_addr));
}

size_t VpnPeer::write(const char* data, size_t len) {
    ssize_t nsent;

    nsent = sendto(_vpn->get_socket_fd(), data, len, MSG_CONFIRM,
            (const struct sockaddr*) &_ext_addr, sizeof(_ext_addr));
    if(nsent < 0)
        throw NetError("Could not send UDP packet", errno, true);

    return (size_t) nsent;
}

size_t VpnPeer::write(const VpnPacket& packet) {
    return write(packet.get_data(), packet.get_data_size());
}

void VpnPeer::got_inbound_packet(const VpnPacket& packet) {
    _packet_loss.log_packet(packet.get_seqno());
}

bool VpnPeer::send_control_packet() {
    if(_rtt.update_due(!_next_control_packet->is_empty()))
        VpnTlvRTTQ::create(*_next_control_packet);

    if(!_next_control_packet->is_empty()) {
        _next_control_packet->prepare_for_sending();
        _vpn->transmit_to_peer(*_next_control_packet);
        cycle_next_control();
        return true;
    }
    return false;
}

void VpnPeer::make_rtta_for(const VpnTlvRTTQ& rttq) {
    VpnTlvRTTA rtta = VpnTlvRTTA::create(*_next_control_packet);
    rtta.set_exp_ts(rttq.get_packet().get_sending_timestamp());
    rtta.set_recv_ts(rttq.get_packet().get_reception_timestamp());
}

PacketLossLogger::PacketLossLogger() : _cur_seqno(0) {}

void PacketLossLogger::log_packet(uint32_t seqno) {
    uint32_t m_seqno = seqno % PACKET_LOSS_HISTSIZE;
    int64_t diff = (int64_t)seqno - _cur_seqno;

    if(diff == 1) {
        _cur_seqno++;
        _packet_loss_hist.reset(m_seqno);
        while(_received_ahead.test((_cur_seqno + 1) % PACKET_LOST_AFTER)) {
            _cur_seqno++;
            _packet_loss_hist.reset(_cur_seqno % PACKET_LOSS_HISTSIZE);
            _received_ahead.reset(_cur_seqno % PACKET_LOST_AFTER);
        }
    } else if(LIKELY(diff > 1)) {
        if(diff < PACKET_LOST_AFTER)
            _received_ahead.set(seqno % PACKET_LOST_AFTER);
        else if(diff < PACKET_LOSS_HISTSIZE) {
            // Packet too much forwards -- consider _cur_seqno lost
            for(int offs=1; offs < PACKET_LOST_AFTER; ++offs) {
                _packet_loss_hist[(_cur_seqno + offs) % PACKET_LOSS_HISTSIZE] =
                    !_received_ahead[(_cur_seqno + offs) % PACKET_LOST_AFTER];
            }
            _received_ahead.reset();
            _cur_seqno = seqno;
            _packet_loss_hist.reset(m_seqno);
        } else
            reboot(); // This is a huge gap -- reboot
    } else {
        if(diff < - 2*PACKET_LOSS_HISTSIZE)
            reboot(); // this is too much backwards -- something's wrong, reboot
        // else: ignore, we've moved forward and counted the packet as lost
    }
}

void PacketLossLogger::reboot() {
    _packet_loss_hist.reset();
    _received_ahead.reset();
}

RTTLogger::RTTLogger() :
    _avg_rtt(0), _cur_rtt(0),
    _last_update(std::chrono::steady_clock::time_point::min())
{
    _last_update =
        std::chrono::steady_clock::now() - std::chrono::seconds(1);
    // This avoids situations where both peers try to renew their RTT at the
    // same time.
    _update_delay = BASE_UPDATE_DELAY + rand() % 100 - 50;
}

void RTTLogger::log(const VpnTlvRTTA& rtt_answer) {
    uint32_t local_t0 = rtt_answer.get_exp_ts(),
             remote_t0 = rtt_answer.get_recv_ts(),
             remote_t1 = rtt_answer.get_packet().get_sending_timestamp(),
             local_t1 = rtt_answer.get_packet().get_reception_timestamp();

    _cur_rtt = (local_t1 - local_t0) - (remote_t1 - remote_t0);
    if(UNLIKELY(_avg_rtt == 0))
        _avg_rtt = _cur_rtt;
    else
        _avg_rtt = (1 - EXP_AVG_FACTOR) * _avg_rtt
            + EXP_AVG_FACTOR * _cur_rtt;

    _last_update = std::chrono::steady_clock::now();
}

bool RTTLogger::update_due(bool soon) const {
    unsigned int ms_since_last_update =
        std::chrono::duration_cast<std::chrono::milliseconds>(
                std::chrono::steady_clock::now() - _last_update).count();
    return ms_since_last_update >= (soon ? _update_delay/2 : _update_delay);
}
